%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%                    The CIFF Proof Procedure                     %%%
%%%        for Abductive Logic Programming with Constraints         %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Constraint projection (file ciff-constraints_result.pl)         %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% v3.0: 23/06/2005 by Markos Hatzitaskos                          %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% This file calculates the "projection" of a given conjunction of 
%%% constraints C w.r.t. a set of variables V.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% project_constraints( +Constraints, +ProjectionVars, -NConstraints).
%%% Top level predicate to project constraints Constraints w.r.t a list
%%% of variables ProjectionVars. NConstraints is a list of constraints
%%% that denotes the projection.
%%% simplify_answer/5 is defined in ``ciff-constraints_sim.pl''
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% project_constraints([B#=<C, B#=<C-1],[B,C],L), simplify_constraints(L, NL).

project_constraints( [], _, []) :- !.
project_constraints( Constraints, ProjectionVars, NConstraints) :-
	simplify_answer( Constraints, TermVars, TermVarsCopy, Template, Body),
	list_answer( TermVars, TermVarsCopy, Body, Template, ResultsList),
	enumerationList( ResultsList, EnumerationList),
	finalConstraintsList( ResultsList, EnumerationList, ProjectionVars, CL2),
	combineLists( CL2, CL3),
	remove_doubles( CL3, CL4),
	remove_numbers( CL4, CL5),
	divide_constraints( CL5, CL6),
	remove_same_dividor( CL6, CL7),
	convert_to_constraints( CL7, NConstraints), !.

remove_same_dividor( [], []).
remove_same_dividor( [(_,C,RHS1)|T1], [([],C,RHS2)|T2]) :-
	remove_same_dividor( T1, T2),
	(
		have_same_dividor( RHS1, _, RHS3) ->
		RHS2 = RHS3
		;
		RHS2 = RHS1
	).
	
have_same_dividor( [], _, []).
have_same_dividor( [(Num,Var)|T1], N, [(Divident,Var)|T2]) :-
	Num = Divident/Dividor,
	N = Dividor,
	have_same_dividor( T1, N, T2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% list_answer( +TermVars, +TermVarsCopy, +Body, +Template, -NConstraints).
%%% Turns the elements in the list of constraints into the following 
%%% form: (ListL,Constraint,ListR). ListL represents the
%%% LHS of the constraint. The Constraint is one of #=<,#>=,#=,#\=. 
%%% ListR represents the RHS of the constraint. The elements in 
%%% these two lists are of the form (Num,Var), which denotes Num*Var,
%%% and the all the elements in these lists are assumed to be added
%%% together.
%%%
%%% termList/2, non_equalities/3, change_var_names/6, change_var_names2/4
%%% are defined in ``ciff-constraints_sim.pl''.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

list_answer( TermVars, TermVarsCopy, Body, Template, NConstraints) :-
	transform_equalities( TermVars, TermVarsCopy, Equalities),
	termList( Body, BodyList),
	non_equalities( BodyList, BodyList1, BodyList2),
	change_var_names( TermVars, Template, BodyList1, 0, BodyList1, BodyList3),
	change_var_names2( TermVars, Template, BodyList2, BodyList4),
	append( BodyList3, BodyList4, BodyList5),
	transform_constraints( BodyList5, Constraints),
	append( Equalities, Constraints, NConstraints).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% transform_equalities( +TermVars, +TermVarsCopy, -Equalities).
%%% If a variable equals a number we add the following to our list of
%%% constraints: ([], #=, [(-1,Var),(Num,1)]).
%%% Var denotes the variable and Num the number the variable equals to.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
transform_equalities( [], [], []).
transform_equalities( [Var|T1], [Num|T2], [([], #=, [(-1,Var),(Num,1)])|T3]) :-
	number( Num), 
	transform_equalities( T1, T2, T3).
transform_equalities( [_|T1], [Var2|T2], T3) :-
	\+ number( Var2), 
	transform_equalities( T1, T2, T3).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% transform_constraints( +Body, -Constraints).
%%% Converts the Body, which had been turned into a list, into a list of 
%%% constraints explained in list_answer/5.
%%%
%%% NOTE: we do not include clpfd:'x*x=y', clpfd:'x*y=z' and clpfd:'x/y=z'
%%% because they will need division. Since in CIFF we will rarely (never?)  
%%% use multiplication or division of variables, we do not get
%%% clpfd:'x*x=y', clpfd:'x*y=z' and clpfd:'x/y=z', and it is safe to
%%% leave them out.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

transform_constraints( [], []).
transform_constraints( [Y in_set L|T1], C) :- !,
	add_Constraints( Y, L, Cn),
	transform_constraints( T1, Cn2),
	append( Cn, Cn2, C).	
transform_constraints( [clpfd:'t+u=<c'(X,Y,Z)|T], C) :- !,
	transform_constraints( T, C1),
	append( [([], #=<, [(-1,X),(-1,Y),(Z,1)])], C1, C).
transform_constraints( [clpfd:'t+u>=c'(X,Y,Z)|T], C) :- !,
	transform_constraints( T, C1),
	append( [([], #>=, [(-1,X),(-1,Y),(Z,1)])], C1, C).		
transform_constraints( [clpfd:'t=<u+c'(X,Y,Z)|T], C) :- !,
	transform_constraints( T, C1),
	append( [([], #=<, [(-1,X),(1,Y),(Z,1)])], C1, C).
transform_constraints( [clpfd:'t>=u+c'(X,Y,Z)|T], C) :- !,
	transform_constraints( T, C1),
	append( [([], #>=, [(-1,X),(1,Y),(Z,1)])], C1, C).
transform_constraints( [clpfd:'t=u+c'(X,Y,Z)|T], C) :- !,
	transform_constraints( T, C1),
	append( [([], #=, [(-1,X),(1,Y),(Z,1)])], C1, C).
transform_constraints( [clpfd:scalar_product(X,Y,#=<,Z)|T], C) :- !,
	transform_constraints( T, C1),
	multiply_Lists( X, Y, XY),
	append( XY, [(Z,1)], XYZ),
	append( [([], #=<, XYZ)], C1, C).
transform_constraints( [clpfd:scalar_product(X,Y,#>=,Z)|T], C) :- !,
	transform_constraints( T, C1),
	multiply_Lists( X, Y, XY),
	append( XY, [(Z,1)], XYZ),
	append( [([], #>=, XYZ)], C1, C).
transform_constraints( [clpfd:'x=<y IND'(X,Y)|T], C) :- !,
	transform_constraints( T, C1),
	append( [([], #=<, [(-1,X),(1,Y)])], C1, C).
transform_constraints( [clpfd:'x>=y IND'(X,Y)|T], C) :- !,
	transform_constraints( T, C1),
	append( [([], #>=, [(-1,X),(1,Y)])], C1, C).
transform_constraints( [clpfd:'t=u IND'(X,Y)|T], C) :- !,
	transform_constraints( T, C1),
	append( [([], #=, [(-1,X),(1,Y)])], C1, C).
transform_constraints( [clpfd:'x+y=t'(X,Y,Z)|T], C) :- !,
	transform_constraints( T, C1),
	append( [([], #=, [(-1,X),(-1,Y),(1,Z)])], C1, C).
transform_constraints( [clpfd:'x+y+z=c'(X,Y,Z,N)|T], C) :- !,
	transform_constraints( T, C1),
	append( [([], #=, [(-1,X),(-1,Y),(-1,Z),(N,1)])], C1, C).
transform_constraints( [clpfd:'x+y+c=z'(X,Y,N,Z)|T], C) :- !,
	transform_constraints( T, C1),
	append( [([], #=, [(-1,X),(-1,Y),(-N,1),(1,Z)])], C1, C).
transform_constraints( [clpfd:'ax+y=t'(N1,X,0,Y)|T], C) :- !,
	transform_constraints( T, C1),
	append( [([], #=, [(-N1,X),(1,Y)])], C1, C).
transform_constraints( [clpfd:'ax+y=t'(N1,X,N2,Y)|T], C) :- !,
	transform_constraints( T, C1),
	append( [([], #=, [(-N1,X),(-N2,1),(1,Y)])], C1, C).
transform_constraints( [clpfd:'t\\=u+c'(_,_,_)|T], C) :- !,
	transform_constraints( T, C).
transform_constraints( [clpfd:'x\\=y IND'(_,_)|T], C) :- !,
	transform_constraints( T, C).
transform_constraints( [clpfd:scalar_product(_,_,#\=,_)|T], C) :- !,
	transform_constraints( T, C).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% enumerationList( +ResultsList, -EnumerationList).
%%% The EnumerationList is created from the ResultsList by including
%%% all the possible constraints that can be created by rotating the
%%% variables in the ResultsList.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

enumerationList( [], []).
enumerationList( [H1|T1], H2) :-
	enumerationList2( H1, 0, List),
	enumerationList( T1, T2),
	append( List, T2, H2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% enumerationList2( +Constraint, +Length, -List).
%%% Creates a List of all the possible constraints that can be created
%%% by rotating the variables in Constraint.
%%% Length starts at 0 and increments everytime by 1. The predicate
%%% stops when Length equals the length of the RHS of the Constraint
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

enumerationList2( ([], _, RHS), N, []) :- length( RHS, N), !.
enumerationList2( ([], Op, RHS), N, [([(Num1, Var)], Op1, RHS1)|Rest]) :-
	nth0(N, RHS, (Num, Var)),
	\+number(Var),
	(
		Num < 0 ->
		Num1 is abs(Num),
		Op1 = Op,
		select( (Num, Var1), RHS, RHS1),
		Var == Var1
		;
		changeOp( Op, Op1),
		Num1 is Num,
		select( (Num, Var1), RHS, RHS2),
		Var == Var1,
		negate_elements( RHS2, RHS1)
	),
	Nn is N+1,
	enumerationList2( ([], Op, RHS), Nn, Rest).
enumerationList2( ([], Op, RHS), N, Rest) :-
	nth0(N, RHS, (_, Var)),
	number(Var),
	Nn is N+1,
	enumerationList2( ([], Op, RHS), Nn, Rest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% finalConstraintsList( +RL, +EL, +PV, -CL).
%%% Calls finalConstraintsList2/5 for every constraint in RL and 
%%% appends the result everytime to CL.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

finalConstraintsList( [], _, _, []).
finalConstraintsList( [H|T], EL, PV, FC) :-
	finalConstraintsList( T, EL, PV, FC1),
	(
		setof(CL,finalConstraintsList2( H, EL, PV, [], CL),ConstraintList) ->
		append( ConstraintList, FC1, FC)
		;
		append( [], FC1, FC)
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% finalConstraintsList2( +Elem, +EL, +PV, +LL, -CL).
%%% We build part of our ConstraintList by replacing variables in  
%%% constraint Elem, which is an element of the ResultsList.
%%% Replacing variables stops whenever only the appropriate variables
%%% remain in the constraint. These variables are given in PV, which is
%%% short for ProjectionVariables, and is a list of variables.
%%% Replacing variables in a constraint means that we substitute
%%% variables, which are not the ones we want to keep in the end, in
%%% the constraint with terms from appropriate constraints in the 
%%% enumeration list EL.
%%% When replacing variables, if the variable is on the RHS and
%%% positive then we search for a constraint in the EL with the same
%%% variable name and predicate. If the variable is on the LHS and
%%% positive we search for a constraint in the EL list with the same
%%% variable name but opposite predicate. If the predicate is equals,
%%% if the replacing variable is on the RHS and positive then we change
%%% the predicate of our equality to the one of the constraint and 
%%% substitute the variable. If the replacing variable is on the LHS
%%% and positive then we change the predicate of our equality to the
%%% inverse of the constraint and substitute the variable.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

finalConstraintsList2( ([], Op, RHS), _, PV, _, [([], Op, RHS)]) :-
	containsOnlyPVs( RHS, PV).
finalConstraintsList2( ([], Op, RHS), EL, PV, LL, FC) :-
	select( (Num, Var), RHS, RHS1),
	\+number(Var),
	\+memberVars( Var, PV),
	member( ([(Num1, Var1)], Op1, RHS2), EL),
	Var==Var1,	%trace,
	divide_by( Num1, RHS2, RHS3), 	
	(				
		Num>0 ->
		( 
			Op1 == Op ->
			Op2 = Op1,
			Op3 = Op
		;
			Op == '#=',
			Op3 = Op1
		),
		append( RHS3, RHS1, RHS5)
		;
		changeOp2( Op, Op1, Op2), 
		(
			Op1 == Op2 ->
			Op3 = Op,
			true
			;
			Op == '#=',
			Op3 = Op2
		),
		negate_elements( RHS3, RHS4),
		append( RHS4, RHS1, RHS5)
	),
	calculate( RHS5, RHS6),
	calculate2( RHS6, RHS7),
	calculate3( RHS7, RHS8),
	check_for_zero(RHS8, RHS9),
	RHS9 \== [], 		% this is when one has substituted with the same constraint
	\+memberLL(([], Op3, RHS9),LL),
	append( [([], Op3, RHS9)], LL, LL1),
	finalConstraintsList2( ([], Op3, RHS9), EL, PV, LL1, FC).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% memberLL( +Elem, +LL).	
%%% Checks if Elem is a member of list LL
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

memberLL(_,[]) :- false.
memberLL(([], Op1, RHS1),[([], Op2, RHS2)|_]) :-
	Op1 == Op2,
	RHS1 == RHS2.
memberLL(([], Op1, RHS1),[_|T]) :-
	memberLL(([], Op1, RHS1),T).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% divide_by( +Num, +List, -NList).
%%% Turns the List that has elements of the form (X, Y) to the list
%%% NList that has elements of the form (X/N, Y). 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

divide_by( _, [], []).
divide_by( 1, T, T) :- !.
divide_by( N, [(Num, Var)|T1], [(Num/N, Var)|T2]) :- !,
	divide_by( N, T1, T2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% check_for_zero( +List, -NList).
%%% Removes elements of the form (0,_) from list List to produce NList.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

check_for_zero( [], []).
check_for_zero( [(0,_)|T1], T2) :-
	check_for_zero( T1, T2).
check_for_zero( [(Num,Var)|T1], [(Num,Var)|T2]) :-
	Num\==0,
	check_for_zero( T1, T2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% calculate( +List, -NList).
%%% Replaces the elements in List of the form (Num1/Num2, Var1) and
%%% (Num1, Var1) where Var1 is a number to (Num/Num2, 1) and (Num, 1)
%%% respectively, where Num equals (Num1*Var1).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

calculate( [], []).
calculate( [(Num1/Num2, Var1)|T1], [(Num/Num2, 1)|T2]) :-
	number( Var1), !,
	Num is (Num1*Var1),
	calculate( T1, T2).
calculate( [(Num1, Var1)|T1], [(Num, 1)|T2]) :-
	number( Var1), !,
	Num is (Num1*Var1),
	calculate( T1, T2).
calculate( [X|T1], [X|T2]) :-
	calculate( T1, T2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% calculate2( +List, -NList).
%%% Adds the elements that can be added together in List and produces
%%% NList. Is needed for division.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

calculate2( [], []).
calculate2( [(Num1/Div, Var1)|T1], T2) :-
	select( (Num2/Div, Var2), T1, T1n),
	number( Var1),
	number( Var2),!,
	Num is (Num1*Var1)+(Num2*Var2), 
	calculate2( [(Num/Div,1)|T1n], T2).
calculate2( [(Num1/Div1, Var1)|T1], T2) :-
	select( (Num2/Div2, Var2), T1, T1n),
	Div1\==Div2,
	number( Var1),
	number( Var2),!,
	Num3 is (Num1*Var1*Div2)+(Num2*Var2*Div1), 
	Div3 is Div1*Div2,
	GCD is gcd(Num3, Div3),
	Num4 is Num3/GCD,
	Div4 is Div3/GCD,
	Num is integer(Num4),
	Div is integer(Div4),
	calculate2( [(Num/Div,1)|T1n], T2).
calculate2( [(Num1/Div, Var1)|T1], T2) :-
	select( (Num2, Var2), T1, T1n),
	Var1 == Var2,
	Num is Num1+(Div*Num2), !,
	calculate2( [(Num/Div,Var1)|T1n], T2).
calculate2( [(Num1, Var1)|T1], T2) :-
	select( (Num2/Div, Var2), T1, T1n),
	Var1 == Var2,
	Num is (Div*Num1)+Num2, !,
	calculate2( [(Num/Div,Var1)|T1n], T2).
calculate2( [X|T1], [X|T2]) :-
	calculate2( T1, T2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% calculate3( +List, -NList).
%%% Adds the elements that can be added together in List and produces
%%% NList.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

calculate3( [], []).
calculate3( [(Num1, Var1)|T1], T2) :-
	select( (Num2, Var2), T1, T1n),
	number( Var1),
	number( Var2),!,
	Num is (Num1*Var1)+(Num2*Var2), 
	calculate3( [(Num,1)|T1n], T2).
calculate3( [(Num1, Var1)|T1], T2) :-
	select( (Num2, Var2), T1, T1n),
	Var1 == Var2,
	Num is Num1+Num2, !,
	calculate3( [(Num,Var1)|T1n], T2).
calculate3( [X|T1], [X|T2]) :-
	calculate3( T1, T2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% containsOnlyPVs( +List, +PV).
%%% Checks if every element in List has only variables from the 
%%% projection variables, PV.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

containsOnlyPVs( [], _).
containsOnlyPVs( [(_, Var)|Rest], PV) :-
	memberVars( Var, PV),
	containsOnlyPVs( Rest, PV).
containsOnlyPVs( [(_, Var)|Rest], PV) :-
	number( Var),
	containsOnlyPVs( Rest, PV).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% memberVars( +Var, +PV).
%%% Checks if variable Var is in the list PV. Could not use build-in
%%% member/2 because Var is a variable and gets instantiated to 
%%% anything in PV.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
memberVars( _, []) :- false. 
memberVars( Var1, [Var2|_]) :- 
	Var1 == Var2.
memberVars( Var1, [Var2|Tail]) :- 
	Var1 \== Var2, 
	memberVars( Var1, Tail).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% changeOp( +Op1, -Op2).
%%% Changes #=< to #>= and #>= to #=<. #= and #\= stay the same.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

changeOp( #=, #=).
changeOp( #\=, #\=).
changeOp( #=<, #>=).
changeOp( #>=, #=<).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% changeOp( +Op1, +Op2, -Op3).
%%% Changes #=< to #>= and #>= to #=<.
%%% If Op1 is #= then if Op2 is #>=, Op3 is #=<.
%%% If Op1 is #= then if Op2 is #=<, Op3 is #>=.
%%% If Op1 is #= then if Op2 is #=, Op3 is #=<.
%%% This is used when replacing variables and the predicate is #=.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

changeOp2( #=, #>=, #=<).
changeOp2( #=, #=<, #>=).
changeOp2( #=, #=, #=).
changeOp2( #=<,#>=, #>=).
changeOp2( #>=,#=<, #=<).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% negate_elements( +List, -NList).
%%% List has elements of the form (Num, Var) where Num is a number.
%%% This predicate returns NList which has the same elements as with
%%% List but with Num negated.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
negate_elements( [], []).
negate_elements( [(Num/Div, Var)|T1], [(NegNumInteger/Div, Var)|T2]) :-
	NegNum is -Num,
	NegNumInteger is integer( NegNum), !,
	negate_elements( T1, T2).
negate_elements( [(Num, Var)|T1], [(NegNumInteger, Var)|T2]) :-
	NegNum is -Num,
	NegNumInteger is integer( NegNum),
	negate_elements( T1, T2).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% multiply_Lists( +List1, +List2, -NList).
%%% Take element H1 from List and H2 from List2 and create NList with 
%%% elements (H3, H2) where H3 is H1 negated.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

multiply_Lists( [], [], []).
multiply_Lists( [H1|T1], [H2|T2], [(H3,H2)|T3]) :- 
	H3 is -H1,
	multiply_Lists( T1, T2, T3).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% add_Constraints( +Constraint, +BodyConstraintList, -ConstraintList).
%%% ConstraintList is created using the Constraint and limiting it into
%%% the range given by the BodyConstraintList.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
add_Constraints( _, [], []).
add_Constraints( _, [[inf|sup]], []) :- !.
add_Constraints( Y, [[inf|V]], [([], #=<, [(-1,Y),(V,1)])]) :- !.
add_Constraints( Y, [[V|sup]], [([], #>=, [(-1,Y),(V,1)])]) :- !.
add_Constraints( Y, [[V|V]], [([], #=, [(-1,Y),(V,1)])]).
add_Constraints( Y, [[L|U]], [([], #>=, [(-1,Y),(L,1)]), ([], #=<, [(-1,Y),(U,1)])]) :-
	L\=U.
add_Constraints( Y, [[inf|U1]|T], NotEquals) :-
	last( T, [_|sup]),
	!,
	add_Constraints2( Y, [[inf|U1]|T], NotEquals).
add_Constraints( Y, [[inf|U1]|T], [([], #=<, [(-1,Y),(U2,1)])|NotEquals]) :- 
	last( T, [_|U2]),
	!,
	add_Constraints2( Y, [[inf|U1]|T], NotEquals).
add_Constraints( Y, [[L1|U1]|T], [([], #>=, [(-1,Y),(L1,1)])|NotEquals]) :-
	last( T, [_|sup]),
	!,
	add_Constraints2( Y, [[L1|U1]|T], NotEquals).
add_Constraints( Y, [[L1|U1]|T], [([], #>=, [(-1,Y),(L1,1)]),([], #=<, [(-1,Y),(U2,1)])|NotEquals]) :-
	last( T, [_|U2]),
	add_Constraints2( Y, [[L1|U1]|T], NotEquals).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% addConstraints2( +Constraint, +BodyConstraintList, -ConstraintList).
%%% ConstraintList is created using the Constraint and limiting it into
%%% the range given by the BodyConstraintList. 
%%% addConstraints2/3 is called by addConstraints/3 when the 
%%% BodyConstraintList contains more than one element. This means that
%%% the Constraint has a disjunction of ranges.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

add_Constraints2( _, [], []).
add_Constraints2( _, [_], []).
add_Constraints2( Y, [[_|U1],[L2|U2]|T], Tail) :-
	add_Constraints2( Y, [[L2|U2]|T], Rest),
	make_Not_EqualsList( Y, U1, L2, List),
	append( List, Rest, Tail).	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% makeNotEqualsList( +Constraint, +LowerBound, +UpperBound, +NotEqualList).
%%% This predicate makes a list, NotEqualList, which contains 
%%% UpperBound minus LowerBound elements of the form 
%%% ([], #\=, [(-1,Y),(N,1)]), where N is a number between LowerBound
%%% and UpperBound.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
make_Not_EqualsList( _, L, U, []) :- 
	N is L+1, 
	N==U, !.
make_Not_EqualsList( Y, L, U, [([], #\=, [(-1,Y),(N,1)])|Rest]) :- 
	N is L+1,
	make_Not_EqualsList( Y, N, U, Rest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% combineLists( +List1, -List2).
%%% List 1 is a list of lists and List 2 combine those lists inside the
%%% list to create one list.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

combineLists( [], []).
combineLists( [H1|T1], List) :-
	combineLists( T1, Rest),
	append( H1, Rest, List).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% remove_doubles( +ConstraintsList, -NConstraints).
%%% NConstraints is the same as ConstraintsList without any doublicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

remove_doubles( [], []).
remove_doubles( [([],C,List1)|T1], L) :-
	member( ([],C,List2), T1),
	permutation( List2, List2perm),
	List1 == List2perm, !,
	remove_doubles( T1, L).
remove_doubles( [([],C1,List1)|T1], L) :-
	changeOp(C1,C2),
	member( ([],C2,List2), T1),
	permutation( List2, List2perm),
	negate_elements(List1, List3),
	List3 == List2perm, !,
	remove_doubles( T1, L).
remove_doubles( [([],C,List)|T1], [([],C,List)|L]) :-
	remove_doubles( T1, L).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% remove_numbers( +ConstraintList, -NConstraints).
%%% Removes any constraints of the form [([],#>=,[(_,Num)])] where Num
%%% is a number. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

remove_numbers( [], []).
remove_numbers( [([],_,List)|T1], L) :-
	only_numbers( List), !, 
	remove_numbers( T1, L).
remove_numbers( [([],C,List)|T1], [([],C,List)|L]) :-
	remove_numbers( T1, L).	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% only_numbers( +List).
%%% Succeds if list has elements of the form ( _,Var) and Var is a 
%%% number.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

only_numbers( []).
only_numbers( [( _,Var)|T]) :-
	number( Var),
	only_numbers( T).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% divide_constraints( +Constraints, -NConstraints).
%%% Takes in a list of Constraints, checks if they can be divided, and
%%% simplifies them by dividing with the appropriate number.
%%% Ex.if Constraints = [([],#=<,[(-2,X),(2,1)])]                 
%%% NConstraints will be [([],#=<,[(-1,X),(1,1)])]
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

divide_constraints( Constraints, NConstraints) :-
	divisable( Constraints, N),
	divide_with( Constraints, N, NConstraints).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% divisable( +Constraints, -Numbers).
%%% Numbers is a list of numbers denoting the number by which each 
%%% constraint in Constraints list can be divided by.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
divisable( [], []).
divisable( [(_,_,RHS)|T], L) :-
	divisable( T, L1),
	(
		divisable2( RHS, Num) ->
		append( [Num], L1, L)
		;
		append( [1], L1, L)
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% divisable2( +Constraint, -Number).
%%% Returns the number Number by which all elements in the constraint 
%%% Constaint can be divided by.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

divisable2( [], 1).
divisable2( [(Num1,_)], Num2) :-
	Num2 is abs(Num1).
divisable2( [(Num1,_), (Num2,Var)|T1], Num11) :-
	Num11 is abs(Num1),
	Num22 is abs(Num2),
	Num11 == Num22,
	divisable2( [(Num2,Var)|T1], Num11).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% divide_with( +Constraints, +N, -NConstraints).
%%% NConstraints is Constraints with every element in each constraint
%%% divided by the relevant element from N.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
divide_with( [], _, []).
divide_with( [(_,C,RHS1)|T1], [Num|T2], [([],C,RHS2)|T3]) :-
	divide_with( T1, T2, T3),
	divide_with2( RHS1, Num, RHS2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% divide_with2( +RHS, +N, -NRHS).
%%% NRHS is RHS with every element in it having its number changed to 
%%% the appropriate one.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

divide_with2( [], _, []).
divide_with2( [(Num1,Var)|T1], N, [(Num2,Var)|T3]) :-
	N\=1,
	Num1 > 1,
	Num2 = 1,
	divide_with2( T1, N, T3).
divide_with2( [(Num1,Var)|T1], N, [(Num2,Var)|T3]) :-
	N\=1,
	Num1 < 0,
	Num2 = -1,
	divide_with2( T1, N, T3).
divide_with2( [(Num1,Var)|T1], N, [(Num2,Var)|T3]) :-	
	N == 1,
	Num2 = Num1,
	divide_with2( T1, N, T3).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% convert_to_constraints( +Constraints, -NConstraints).
%%% Turns the elements in the Constraints list, which have the form
%%% ( List1, Constraint, List2), explained in list_answer/5, back into
%%% constraints of one of the following forms: X#=<Y, X#>=Y, X#=Y, 
%%% or X#\=Y. X represents the LHS of the constraint and Y the RHS.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

convert_to_constraints( Constraints, NConstraints) :-
	make_elem_pos( Constraints, C1),
	multiply_tuples( C1, C2),
	add_tuples( C2, C3),
	make_constraints( C3, NConstraints).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% make_elem_pos( +Constraints, -NConstraints).
%%% NConstraints is the same as Constraints but with the constraints
%%% not having any negative elements. If there is some negative element
%%% on one side of the constraint it is made positive and moved to the
%%% the other side.
%%% Ex. [([],#=<,[(-2,X),(2,1)])]  is turned to [([(2,X)],#=<,[(2,1)])] 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

make_elem_pos( [], []).
make_elem_pos( [(_,C,RHS1)|T], L) :-
	make_elem_pos( T, L1),
	make_elem_pos2( RHS1, LHS2, RHS2),
	append( [(LHS2,C,RHS2)], L1, L).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% make_elem_pos2( +RHS, -LHS1, -RHS1).
%%% RHS is a list of elements (Num,Var). All the elements in which 
%%% Num < 0 are made positive and put in LHS1 and all the elements
%%% in which Num >= 1 are put in RHS1.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
make_elem_pos2( [], [], []).
make_elem_pos2( [(Num,Var)|T1], [(NumAbs,Var)|T2], T3) :-
	Num < 0,
	negate_elements( [(Num,Var)], [(NumAbs,Var)]),
	make_elem_pos2( T1, T2, T3).
make_elem_pos2( [(Num,Var)|T1], T2, [(Num,Var)|T3]) :-
	Num >= 1,
	make_elem_pos2( T1, T2, T3).
make_elem_pos2( [(Num,_)|T1], T2, T3) :-
	Num == 0,
	make_elem_pos2( T1, T2, T3).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% multiply_tuples( +Constraints, -NConstraints)
%%% NConstraints is the same as Constraints with elements (Num,Var)
%%% replaced by Num*Var in both the RHS and LHS of every constraint.
%%% Ex. [([(2,X)],#=<,[(2,1)])]  is turned into [([(2*X)],#=<,[(2*1)])] 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

multiply_tuples( [], []).
multiply_tuples( [(LHS1,C,RHS1)|T], L) :-
	multiply_tuples( T, L1), 
	multiply_tuples2( LHS1, LHS2),
	multiply_tuples2( RHS1, RHS2),
	append( [(LHS2,C,RHS2)], L1, L).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% multiply_tuples2( +List, -NList).
%%% List is a list of (Num,Var) elements and NList is a list of
%%% (Num*Var) elements.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

multiply_tuples2( [], []).
multiply_tuples2( [(Num,Var)|T1], [Var|T2]) :-
	\+number(Var),
	Num == 1,
	multiply_tuples2( T1, T2).
multiply_tuples2( [(Num,Var)|T1], [Num*Var|T2]) :-
	\+number(Var),
	Num \= 1,
	multiply_tuples2( T1, T2).
multiply_tuples2( [(Num/Div,Var)|T1], [NumVar/Div|T2]) :-
	number(Var),
	NumVar is Num*Var,
	!,
	multiply_tuples2( T1, T2).
multiply_tuples2( [(Num,Var)|T1], [NumVar|T2]) :-
	number(Var),
	NumVar is Num*Var,
	multiply_tuples2( T1, T2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% add_tuples( +Constraints, -NConstraints),
%%% NConstraints is the same as Constraints with all elements in its
%%% two lists added.
%%% Ex. [([2*X,D,3],#=<,[2])] is turned into
%%% [([2*X+D+3],#=<,[2])]
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
add_tuples( [], []).
add_tuples( [(LHS1,C,RHS1)|T], L) :-
	add_tuples( T, L1), 
	add_tuples2( LHS1, [LHS2]),
	add_tuples2( RHS1, [RHS2]),
	append( [(LHS2,C,RHS2)], L1, L).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% add_tuples2( +List, -NList).
%%% List is a list of elements and NList is a list with one element
%%% which is all the elements of list List added together.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

add_tuples2( [], [0]).
add_tuples2( [NumVar], [NumVar]) :- !.
add_tuples2( [NumVar|T], [NumVar2]) :-
	add_tuples2( T, [Elem]),
	NumVar2 = Elem+NumVar.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% make_constraints( +Constraints, -NConstraints).
%%% Constraints list has constraint elements of the form 
%%% (LHS,Constraint,RHS). NConstraints list has the same constraints
%%% as Constraints list but changes their form to the normal form
%%% LHS Constraint RHS.
%%% Ex [([2*X+D+3],#=<,[2])] is turned into [2*X+D+3#=<2].
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

make_constraints( [], []).
make_constraints( [(LHS,#=,RHS)|T1], [LHS#=RHS|T2]) :-
	make_constraints( T1, T2).
make_constraints( [(LHS,#\=,RHS)|T1], [LHS#\=RHS|T2]) :-
	make_constraints( T1, T2).	
make_constraints( [(LHS,#>=,RHS)|T1], [LHS#>=RHS|T2]) :-
	make_constraints( T1, T2).
make_constraints( [(LHS,#=<,RHS)|T1], [LHS#=<RHS|T2]) :-
	make_constraints( T1, T2).
make_constraints( [(LHS,#>,RHS)|T1], [LHS#>RHS|T2]) :-
	make_constraints( T1, T2).
make_constraints( [(LHS,#<,RHS)|T1], [LHS#<RHS|T2]) :-
	make_constraints( T1, T2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
