%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%                    The CIFF Proof Procedure                     %%%
%%%        for Abductive Logic Programming with Constraints         %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Constraint simplification (file ciff-constraints_sim.pl)        %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% v3.0: 23/06/2005 by Markos Hatzitaskos                          %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% This file simplifies a given list of constraints.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% display_constraints( +Constraints, -NConstraints).
%%% Top level predicate to simplify constraints. Input a list of 
%%% constraints and get another list NConstraints with the constraints
%%% simplified.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

display_constraints( [], []) :- !.
display_constraints( Constraints, NConstraints) :-
	simplify_answer( Constraints, TermVars, TermVarsCopy, Template, Body),
	convert_answer( TermVars, TermVarsCopy, Body, Template, NConstraints), !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% simplify_answer( +C, -TermVars, -TermVarsCopy, -f(TermVars), -Body).
%%% Copies the constraints C into a new list and finds the
%%% term variables TermVars of both the original and the copied
%%% constraint list, TermVarsCopy. Then converts the copied list into
%%% a Term and uses the Term in conjunction with fd_copy_term/3
%%% to simplify the constraints.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

simplify_answer( C, TermVars, TermVarsCopy, Template, []) :-
	copy_term( C, Ccopy),
	term_variables_bag( C, TermVars),
	term_variables_bag( Ccopy, TermVarsCopy),
	listTerm( Ccopy, Term),
	Term,
	fd_copy_term( f( TermVarsCopy), f( Template), true), !.
simplify_answer( C, TermVars, TermVarsCopy, Template, Body) :-
	copy_term(C, Ccopy),
	term_variables_bag( C, TermVars),
	term_variables_bag( Ccopy, TermVarsCopy),
	listTerm( Ccopy, Term),
	Term,
	fd_copy_term( f( TermVarsCopy), f( Template), Body).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% convert_answer( +TermVars, +TermVarsCopy, +Body, -NewConstraints).
%%% Converts the answer given by simplify_answer/5 in order to be 
%%% more readable.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

convert_answer( TermVars, TermVarsCopy, Body, Template, NewConstraints) :-
	equalities( TermVars, TermVarsCopy, Equalities),
	termList( Body, BodyList),
	non_equalities( BodyList, BodyList1, BodyList2),
	change_var_names( TermVars, Template, BodyList1, 0, BodyList1, BodyList3),
	change_var_names2( TermVars, Template, BodyList2, BodyList4),
	append( BodyList3, BodyList4, BodyList5),
	convertToConstraints( BodyList5, BodyList6),
	remove_non_termvars( BodyList6, TermVars, Constraints),
	append( Equalities, Constraints, NewConstraints).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% remove_non_termvars( +Body, +TermVars, -NewBody).
%%% Body is a list of constraints. TermVars is a list of variables.
%%% NewBody is a list of constraints that only have variables that are
%%% elements in TermVars list.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

remove_non_termvars( [], _, []).
remove_non_termvars( [Con|T], TermVars, [Con|Constraints]) :-
	term_variables( Con, ConVars),
	sublist_noOrder( ConVars, TermVars),
	remove_non_termvars( T, TermVars, Constraints).
remove_non_termvars( [Con|T], TermVars, Constraints) :-
	term_variables( Con, ConVars),
	\+sublist_noOrder( ConVars, TermVars),
	remove_non_termvars( T, TermVars, Constraints).	
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% sublist_noOrder( +Variables, +TermVariables).
%%% This predicate unifies to true if all the elements of the Variables
%%% list also exist in the TermVariables list.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
sublist_noOrder( [], _).
sublist_noOrder( [Var1|T], TermVars) :-
	member( Var2, TermVars),
	Var1 == Var2,
	sublist_noOrder( T, TermVars).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% equalities( +TermVars, +TermVarsCopy, -Equalities).
%%% Adds the equality constraints to the new constraints list,
%%% Equalities. 
%%% Equality constraints have the form X#=N where X is a
%%% variable and N is a number.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

equalities( [], [], []).
equalities( [Var|T1], [Num|T2], [Var#=Num|T3]) :-
	number( Num), 
	equalities( T1, T2, T3).
equalities( [_|T1], [Var2|T2], T3) :-
	\+ number( Var2), 
	equalities( T1, T2, T3).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% change_var_names( +TermVars, +Template, +Body, +N, +Body, -NewBody).
%%% This predicate replaces the variable X in the terms X in_set Y of 
%%% Body with the corresponding variable from TermVars. After it has
%%% replaced all the variables it returns a new list NewBody.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

change_var_names( _, [], _, _, _, []).
change_var_names( TermVars, [_|T1], [], N, Body, T3) :-
	N1 is N+1,
	change_var_names( TermVars, T1, Body, N1, Body, T3).
change_var_names( TermVars, [H1|T1], _, N, Body, T3) :-
	number( H1),
	N1 is N+1,
	change_var_names( TermVars,  T1, Body, N1, Body, T3).
change_var_names( TermVars, [H1|T1], [H2 in_set Set|_], N, Body, [TermVar in_set Set|T3]) :- 
	H1 == H2,
	nth0( N, TermVars, TermVar),
	N1 is N+1,
	change_var_names( TermVars, T1, Body, N1, Body, T3).
change_var_names( TermVars, [H1|T1], [H2 in_set _|T2], N, Body, T3) :-
	H1 \== H2,
	change_var_names( TermVars, [H1|T1], T2, N, Body, T3).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% change_var_names2( +TermVars, +Template, +Body, -NewBody).
%%% This predicate takes in a list of elements of the forms:
%%%
%%% clpfd:'t=<u+c'(X,Y,Z), clpfd:'t>=u+c'(X,Y,Z), clpfd:'t=u+c'(X,Y,Z),
%%% clpfd:scalar_product(X,Y,#=<,Z), clpfd:scalar_product(X,Y,#>=,Z),
%%% clpfd:'x=<y IND'(X,Y), clpfd:'x>=y IND'(X,Y), clpfd:'t=u IND'(X,Y), 
%%% clpfd:'x+y=t'(X,Y,Z), clpfd:'x+y+z=c'(X,Y,Z,N), clpfd:'x+y+c=z'(X,Y,N,Z), 
%%% clpfd:'ax+y=t'(N1,X,0,Y), clpfd:'ax+y=t'(N1,X,N2,Y), 
%%% clpfd:'x*x=y'(X,Y), clpfd:'x*y=z'(X,Y,Z), clpfd:'x/y=z'(X,Y,Z)
%%% clpfd:'t\\=u+c'(X,Y,Z), clpfd:'x\\=y IND'(X,Y),
%%% clpfd:scalar_product(X,Y,#\=,Z), clpfd:'t+u>=c'(X,Y,Z),
%%% clpfd:'t+u>=c'(X,Y,Z)
%%%
%%% and replaces the variables in those elements with the corresponding
%%% ones from TermVars. NewBody is the same list as Body with the 
%%% variables replaced.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

change_var_names2( _, _, [], []) :- !.
change_var_names2( TermVars, Template, Body, NewBody) :-
	remove_numbers( TermVars, Template, NTermVars, NTemplate),
	subst( NTemplate, NTermVars, Body, NewBody).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% remove_numbers( +TermVars, +Template, -NTermVars, -NTemplate).
%%% Takes in a two lists, TermVars and Template and returns two new 
%%% lists NTermVars and NTemplate, with the same elements minus the
%%% number elements in NTemplate and the corresponding element in 
%%% NTermVars.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

remove_numbers( [], [], [], []).
remove_numbers( [_|T1], [H2|T2], T3, T4) :-
	number( H2), 
	remove_numbers( T1, T2, T3, T4).
remove_numbers( [H1|T1], [H2|T2], [H1|T3], [H2|T4]) :-
	\+number( H2), 
	remove_numbers( T1, T2, T3, T4).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% non_equalities( +Body, -BodyList1, -BodyList2),
%%% This predicate receives a list of constraints Body and makes a list,
%%% BodyList1, of those constraints in Body that are of the form 
%%% X in_set Y. 
%%% It also makes another list, BodyList2, of the rest of the 
%%% constraints in Body.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

non_equalities( [], [], []).
non_equalities( [X in_set S|T1], [X in_set S|T2], T3) :-
	!, non_equalities( T1, T2, T3).
non_equalities( [X|T1], T2, [X|T3]) :-
	!, non_equalities( T1, T2, T3).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% listTerm( +List, -Term).
%%% Convert a List into a Term.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

listTerm( [],true).
listTerm( [X|Xs],Tuple) :- listTerm_( Xs,X,Tuple).
listTerm_( [],X,X).
listTerm_( [Y|Ys],X,(X,Tuple)) :- listTerm_( Ys,Y,Tuple).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%	
%%% termList( +Term, -List).
%%% Convert a Term into a List.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

termList( [], []).
termList( (X,Tuple), [X|L]) :- !, termList( Tuple, L).
termList( X, [X]).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% convertToConstraints( +BodyList, -Constraints).
%%% Converts the Body received from fd_copy_term/3, which had been 
%%% turned into a list, into a list of simple Constraints.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

convertToConstraints( [], []).
convertToConstraints( [Y in_set L|T1], C) :- !,
	addConstraints( Y, L, Cn),
	convertToConstraints( T1, Cn2),
	append( Cn, Cn2, C).
convertToConstraints( [clpfd:'t=<u+c'(X,Y,Z)|T], C) :- !,
	convertToConstraints( T, C1),
	(
		Z<0 ->
		Zn is abs(Z),
		append( [X#=<Y-Zn], C1, C)
		;
		append( [X#=<Y+Z], C1, C)
	).
convertToConstraints( [clpfd:'t>=u+c'(X,Y,Z)|T], C) :- !,
	convertToConstraints( T, C1),
	(
		Z<0 ->
		Zn is abs(Z),
		append( [X#>=Y-Zn], C1, C)
		;
		append( [X#>=Y+Z], C1, C)
	).

convertToConstraints( [clpfd:'t+u=<c'(X,Y,Z)|T], C) :- !,
	convertToConstraints( T, C1),
	(
		Z<0 ->
		Zn is abs(Z),
		append( [X+Y+Zn#=<0], C1, C)
		;
		append( [X+Y#=<Z], C1, C)
	).
convertToConstraints( [clpfd:'t+u>=c'(X,Y,Z)|T], C) :- !,
	convertToConstraints( T, C1),
	(
		Z<0 ->
		Zn is abs(Z),
		append( [X+Y+Zn#>=0], C1, C)
		;
		append( [X+Y#>=Z], C1, C)
	).
convertToConstraints( [clpfd:'t=u+c'(X,Y,Z)|T], C) :- !,
	convertToConstraints( T, C1),
	(
		Z<0 ->
		Zn is abs(Z),
		append( [X#=Y-Zn], C1, C)
		;
		append( [X#=Y+Z], C1, C)
	).
convertToConstraints( [clpfd:scalar_product(X,Y,#=<,Z)|T], C) :- !,
	convertToConstraints( T, C1),
	multiplyLists( X, Y, XY1, XY2),
	sumConstraints( XY1, XY11),
	sumConstraints( XY2, XY22),
	(
		Z>0 ->
		append( [XY11#=<XY22+Z], C1, C)
		;
		Zn is abs( Z),
		append( [XY11#=<XY22-Zn], C1, C)
	).
convertToConstraints( [clpfd:scalar_product(X,Y,#>=,Z)|T], C) :- !,
	convertToConstraints( T, C1),
	multiplyLists( X, Y, XY1, XY2),
	sumConstraints( XY1, XY11),
	sumConstraints( XY2, XY22),
	(
		Z>0 ->
		append( [XY11#>=XY22+Z], C1, C)
		;
		Zn is abs( Z),
		append( [XY11#>=XY22-Zn], C1, C)
	).
convertToConstraints( [clpfd:'x=<y IND'(X,Y)|T], C) :- !,
	convertToConstraints( T, C1),
	append( [X#=<Y], C1, C).
convertToConstraints( [clpfd:'x>=y IND'(X,Y)|T], C) :- !,
	convertToConstraints( T, C1),
	append( [X#>=Y], C1, C).
convertToConstraints( [clpfd:'t=u IND'(X,Y)|T], C) :- !,
	convertToConstraints( T, C1),
	append( [X#=Y], C1, C).	
convertToConstraints( [clpfd:'x+y=t'(X,Y,Z)|T], C) :- !,
	convertToConstraints( T, C1),
	append( [X+Y#=Z], C1, C).
convertToConstraints( [clpfd:'x+y+z=c'(X,Y,Z,N)|T], C) :- !,
	convertToConstraints( T, C1),
	(
		N<0 ->
		Nn is abs(N),
		append( [X+Y+Z+Nn#=0], C1, C)
		;
		append( [X+Y+Z#=N], C1, C)
	).
convertToConstraints( [clpfd:'x+y+c=z'(X,Y,N,Z)|T], C) :- !,
	convertToConstraints( T, C1),
	(
		N<0 ->
		Nn is abs(N),
		append( [X+Y#=Nn+Z], C1, C)
		;
		append( [X+Y+N#=Z], C1, C)	
	).
convertToConstraints( [clpfd:'ax+y=t'(N1,X,0,Y)|T], C) :- !,
	convertToConstraints( T, C1),
	(
		N1<0 ->
		N1n is abs(N1),
		append( [0#=Y+N1n*X], C1, C)
		;	
		append( [N1*X#=Y], C1, C)
	).
convertToConstraints( [clpfd:'ax+y=t'(N1,X,N2,Y)|T], C) :- !,
	convertToConstraints( T, C1),
	(
		N1<0 ->
		N1n is abs(N1),
		(
			N2<0 ->
			N2n is abs(N2),
			append( [0#=Y+N1n*X+N2n], C1, C)
			;
			append( [N2#=Y+N1n*X], C1, C)
		)
		;
		(
			N2<0 ->
			N2n is abs(N2),
			append( [N1*X#=Y+N2n], C1, C)
			;
			append( [N1*X+N2#=Y], C1, C)
		)
	).
convertToConstraints( [clpfd:'x*x=y'(X,Y)|T], C) :- !,
	convertToConstraints( T, C1),
	append( [X*X#=Y], C1, C).
convertToConstraints( [clpfd:'x*y=z'(X,Y,Z)|T], C) :- !,
	convertToConstraints( T, C1),
	append( [X*Y#=Z], C1, C).
convertToConstraints( [clpfd:'x/y=z'(X,Y,Z)|T], C) :- !,
	convertToConstraints( T, C1),
	append( [X/Y#=Z], C1, C).
convertToConstraints( [clpfd:'x\\=y IND'(X,Y)|T], C) :- !,
	convertToConstraints( T, C1),
	append( [X#\=Y], C1, C).
convertToConstraints( [clpfd:'t\\=u+c'(X,Y,Z)|T], C) :- !,
	convertToConstraints( T, C1),
	(
		Z<0 ->
		Zn is abs(Z),
		append( [X#\=Y-Zn], C1, C)
		;
		append( [X#\=Y+Z], C1, C)
	).
convertToConstraints( [clpfd:scalar_product(X,Y,#\=,Z)|T], C) :- !,
	convertToConstraints( T, C1),
	multiplyLists( X, Y, XY1, XY2),
	sumConstraints( XY1, XY11),
	sumConstraints( XY2, XY22),
	(
		Z>0 ->
		append( [XY11#\=XY22+Z], C1, C)
		;
		Zn is abs( Z),
		append( [XY11#\=XY22-Zn], C1, C)
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% multiplyLists( +X, +Y, -XY1, -XY2).
%%% List X is of the form [A,B,C], where A,B,C are numbers, and list Y
%%% is of the form [D, E, F], where D,E,F are variables. 
%%% multiplyLists/4 returns two lists XY1 and XY2. XY1 contains 
%%% elements of the form [A*D, B*E, etc.], where A,B are positive
%%% numbers, and XY2 contains elements of the form [C*F, etc.], where
%%% C is a negative number.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

multiplyLists( [], [], [], []).
multiplyLists( [H1|T1], [H2|T2], [H1*H2|T3], T4) :- 
	multiplyLists( T1, T2, T3, T4),
	H1>0.
multiplyLists( [H1|T1], [H2|T2], T3, [H11*H2|T4]) :- 
	multiplyLists( T1, T2, T3, T4),
	H1<0,
	H11 is abs( H1).	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% sumConstraints( +List, -Sum).
%%% Add all the elements of List, which are of the form A*B, where
%%% A is a number and B is a variable. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

sumConstraints( [], 0).
sumConstraints( [1*H1], N) :- N = H1, !.
sumConstraints( [V*H1], N) :- V>1, N = (V*H1), !.
sumConstraints( [1*H1|T1], N) :- sumConstraints( T1, N1), N = N1+H1.
sumConstraints( [V*H1|T1], N) :- V>1, sumConstraints( T1, N1), N = N1+(V*H1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% addConstraints( +Constraint, +BodyConstraintList, -ConstraintList).
%%% ConstraintList is created using the Constraint and limiting it into
%%% the range given by the BodyConstraintList.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

addConstraints( _, [], []).
addConstraints( _, [[inf|sup]], []) :- !.
addConstraints( Y, [[inf|V]], [Y#=<V]) :- !.
addConstraints( Y, [[V|sup]], [V#=<Y]) :- !.
addConstraints( Y, [[V|V]], [Y#=V]).
addConstraints( Y, [[L|U]], [L#=<Y, Y#=<U]) :-
	L\=U.
addConstraints( Y, [[inf|U1]|T], NotEquals) :-
	last( T, [_|sup]),
	!,
	addConstraints2( Y, [[inf|U1]|T], NotEquals).
addConstraints( Y, [[inf|U1]|T], [Y#=<U2|NotEquals]) :- 
	last( T, [_|U2]),
	!,
	addConstraints2( Y, [[inf|U1]|T], NotEquals).
addConstraints( Y, [[L1|U1]|T], [L1#=<Y|NotEquals]) :-
	last( T, [_|sup]),
	!,
	addConstraints2( Y, [[L1|U1]|T], NotEquals).
addConstraints( Y, [[L1|U1]|T], [L1#=<Y, Y#=<U2|NotEquals]) :-
	last( T, [_|U2]),
	addConstraints2( Y, [[L1|U1]|T], NotEquals).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% addConstraints2( +Constraint, +BodyConstraintList, -ConstraintList).
%%% ConstraintList is created using the Constraint and limiting it into
%%% the range given by the BodyConstraintList. 
%%% addConstraints2/3 is called by addConstraints/3 when the 
%%% BodyConstraintList contains more than one element. This means that
%%% the Constraint has a disjunction of ranges.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

addConstraints2( _, [], []).
addConstraints2( _, [_], []).
addConstraints2( Y, [[_|U1],[L2|U2]|T], Tail) :-
	addConstraints2( Y, [[L2|U2]|T], Rest),
	makeNotEqualsList( Y, U1, L2, List),
	append( List, Rest, Tail).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% makeNotEqualsList( +Constraint, +LowerBound, +UpperBound, +NotEqualList).
%%% This predicate makes a list, NotEqualList, which contains 
%%% UpperBound minus LowerBound elements of the form Constraint#\=N, 
%%% where N is a number between LowerBound and UpperBound.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

makeNotEqualsList( _, L, U, []) :- 
	N is L+1, 
	N==U, !.
makeNotEqualsList( Y, L, U, [Y#\=N|Rest]) :- 
	N is L+1,
	makeNotEqualsList( Y, N, U, Rest).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
