%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%                    The CIFF Proof Procedure                     %%%
%%%        for Abductive Logic Programming with Constraints         %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% GUI (file ciff-gui.pl)                                          %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% v3.0: 23/06/2005 by Markos Hatzitaskos                          %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% This file, along with `ciff-gui.tcl', provide a simple graphical
%%% user interface for CIFF.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Load Tcl/Tk library.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(library(tcltk)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Define dynamic predicate tcl/1 to hold a reference to the Tcl 
%%% interpreter.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

?- dynamic( tcl/1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% setup.
%%% Load up the GUI and assert the value of the Tcl interpreter into
%%% dynamic predicate tcl.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

setup :- 
    tk_new([name('CIFF')], Tcl),
    debugging_in_gui( on),
    retractall( tcl),
    assert( tcl( Tcl)),
    tcl_eval(Tcl, [source,'ciff-gui.tcl'], _),
    tk_main_loop,
    debugging_in_gui( off).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% submit( +File).
%%% This predicate is called when the `Run CIFF' button is pressed.
%%% It takes File as input, which represents a file that containts 
%%% iff definitions and ICs. It first calls read_alp/3 to read in the
%%% File. It then runs ciff with Defs, ICs, and Query, from the File,
%%% as arguments and retrieves all the answers.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

submit( File, Query ) :-
	tcl( Tcl),
	read_alp( File, Defs, ICs),
	(
		ciff( Defs, ICs, Query, _), fail
		;
		tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,br(write('no'))], _)
	).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% switch_negation_as_failure( +N).
%%% Switches the switch_naf dynamic clause on or off.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

switch_negation_as_failure( 0) :- switch_naf( off).
switch_negation_as_failure( 1) :- switch_naf( on).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% inform_gui( +Rule, +Arguments, +Node). 
%%% Used for debugging purposes. The output depends on the debugging
%%% mode chosen (via switch_debugging/1). The following modes are
%%% available:
%%% "on":    full debugging information, namely print out last Rule 
%%%          applied (with Arguments) and current Node;
%%% "step":  same as "on", but Prolog will wait for the user to hit 
%%%          the return key;
%%% "short": only show the name of the Rule used;
%%% "dots":  only print a single dot for each step;
%%% "stats:  print out statistics at the end (counting takes place
%%%          within inform/3).
%%% "off":   no output [default].
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

inform_gui( Rule, _, _) :-
  Rule \== 'Answer',
  debugging( off), !.

inform_gui( 'Answer', _, A:S:C) :-
  (debugging( on) ; debugging( step) ; debugging( off)), !,	
  tcl( Tcl),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write(\n)], _),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,br(write('Answer = '))], _),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,br(write( A:S:C))], _),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write(\n\n)], _),
  tk_do_all_events,
  tk_next_event(Tcl, next).

inform_gui( Rule, Arguments, Node) :-
  (debugging( on) ; debugging( step)), !,
  tcl( Tcl),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write(\n)], _),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,br(Rule)], _),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,br(write(': '))], _),
  variable_list( Arguments, 'X', NArguments, _),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,br(write(NArguments))], _),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write(\n)], _),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,br(write('node: '))], _), 
  variable_list( Node, 'X', NNode, _), 
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,br(write(NNode))], _),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write(\n\n)], _),
  tk_do_all_events,
  (debugging( step) -> tk_next_event(Tcl, next) ; nl).

inform_gui( Rule, _, _) :-
  debugging( short), !,
  tcl( Tcl),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write(\n)], _), 
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,br(Rule)], _),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write(\n\n)], _).

inform_gui( Rule, _, _) :-
  debugging( stats), !, 
  update_gui_stats( Rule).
  
inform_gui( _, _, _) :-
  tcl( Tcl),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,br(write('.'))], _).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% tk_do_all_events.
%%% It is used to refresh the tcl/tk GUI.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

tk_do_all_events :-
    	tk_do_one_event, !,
    	tk_do_all_events.
tk_do_all_events.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% update_gui_stats( +Rule).
%%% Update the counters used to collect proof statistics. The action
%%% taken depends on the proof rule applied.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

update_gui_stats( 'initialise') :- !,
  init_counter( proof_steps),
  init_counter( branches),
  init_clock.

update_gui_stats( 'splitting') :- !,
  step_counter( proof_steps),
  step_counter( branches),
  restart_clock.

update_gui_stats( 'final constraint checking (success)') :- !,
  step_counter( proof_steps),
  print_gui_stats.

update_gui_stats( _) :-
  step_counter( proof_steps).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% print_gui_stats.
%%% Print out statistics (in case debugging mode "stats" is in use).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

print_gui_stats :- !,
  tcl( Tcl),
  stop_clock,
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write(\n)], _),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,br(write('CIFF Statistics'))], _), 
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write(\n)], _),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,br(write('---------------'))], _),	
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write(\n)], _),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,br(write('Steps:'))], _),
  read_counter( proof_steps, ProofSteps),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write( ProofSteps)], _),	
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write(\n)], _),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,br(write('Branches:'))], _),
  read_counter( branches, BranchesCounted),
  (BranchesCounted = 0 -> Branches = 1 ; Branches = BranchesCounted),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write( Branches)], _),	
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write(\n)], _),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,br(write('Time [ms]:'))], _),
  read_clock( Time),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write( Time)], _), 
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write(\n)], _),
  tcl_eval( Tcl, [dot([main,fra34,cpd48,3]),insert,end,write(\n)], _), !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%