%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%                    The CIFF Proof Procedure                     %%%
%%%        for Abductive Logic Programming with Constraints         %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Negation as failure (file ciff-naf.pl)                          %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% v3.0: 23/06/2004 by Markos Hatzitaskos                          %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% This file provides the implementation of the auxilary definitions
%%% for negation as failure. Use switch_naf/1 to switch the use of
%%% negation as failure on and off (default is off).
%%%
%%% The extension of CIFF with negation as failure is based on the
%%% following paper: F. Sadri and F. Toni. Abduction with Negation 
%%% as Failure for Active and Reactive Rules. Proceedings of the 6th 
%%% Congress of the Italian Association for Artificial Intelligence 
%%% (AI*IA-1999), Springer-Verlag, 2000.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% switch_naf( +Status).
%%% Switch negation as failure on or off. Status should be either "on" 
%%% or "off" (default is "off"). In the former case, CIFF with negation 
%%% as failure is used, otherwise the standard version of CIFF (with
%%% classical negation) will be executed.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

switch_naf( Status) :-
  retractall( naf(_)),
  assert( naf(Status)).

:- switch_naf( off).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% mark_ics( +ICs, -NewICs).
%%% Mark integrity constraints in an ALP with an '*' symbol, provided 
%%% "negation as failure"-switch is on.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

mark_ics( [As implies Bs|ICs], [[*|As] implies Bs|MarkedICs]) :- 
  naf(on), !,
  mark_ics( ICs, MarkedICs).

mark_ics( ICs, ICs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% deSkolemize( +SkolemEqualities, +NSkol, +NSkol, -Constraints)
%%% SkolemEqualities is a list of the skolemisation of free and 
%%% existential quantified variables in the Node.
%%% NSkol is a list of equalities for the skolemised variables
%%% Constraints is a list where we have combined the variables from 
%%% SkolemEqualities to the results from NSkol.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

deSkolemize( [], _, _, []).
deSkolemize( [_|T1], [], T2, T3) :-
	deSkolemize( T1, T2, T2, T3).
deSkolemize( [Skol1=Var|T1], [Skol2=_|T2], T3, T4) :-
	Skol1 \= Skol2,
	deSkolemize( [Skol1=Var|T1], T2, T3, T4).
deSkolemize( [Skol=Var|T1], [Skol=Cons|T2], T3, [Var=Cons|T4]) :-
	deSkolemize( [Skol=Var|T1], T2, T3, T4).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% remove_skolemEqualities( +SkolemEqualities, -NSkolemEqualities).
%%% NSkolemEqualities is the same as SkolemEqualities without any 
%%% elements of the for skol=skol.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

remove_skolemEqualities( [], []).
remove_skolemEqualities( [L=R|T1], [L=R|T2]) :-
	var(R), !,
	remove_skolemEqualities( T1, T2).
remove_skolemEqualities( [L=R|T1], T2) :- 
	name( L, L1), 
	prefix( [115,107,111,108], L1),
	name( R, R1), 
	prefix( [115,107,111,108], R1), !,
	remove_skolemEqualities( T1, T2).
remove_skolemEqualities( [E|T1], [E|T2]) :-
	!, remove_skolemEqualities( T1, T2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% substToEqualities( +Substitutions, -Equalities).
%%% Turns a list with elements of the form A/B to a list with the same
%%% elements but of the form A=B.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

substToEqualities( [], []).
substToEqualities( [A / B|T1], [A = B|T2]) :-
	substToEqualities( T1, T2).
substToEqualities( [not(A / B)|T1], [not(A = B)|T2]) :-
	substToEqualities( T1, T2).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% create_false_list( +List, -MarkedList).
%%% For each [A] implies [false] element in a list a 
%%% [[*, A]implies[false]] element is created.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

create_false_list( [], []).
create_false_list( [[A] implies [false]|Tail], [[*, A] implies [false]|List]) :-
	create_false_list( Tail, List).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% create_provable_list( +List, -MarkedList).
%%% For each [A] implies [false] element in a list a provable([A]) 
%%% element is created.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

create_provable_list( [], []).
create_provable_list( [[A] implies [false] | Tail], [[provable([A])]|List]) :-
	create_provable_list( Tail, List).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% check_marked_implication( +List).
%%% Checks that a list contains only elements of the 
%%% form [_] implies [false].
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

check_marked_implication( []).
check_marked_implication( [[_] implies [false] | Tail]) :-
		check_marked_implication( Tail).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% getAbducibles( +Node, -List).
%%% Retrieves the abducibles from a node.
%%%
%%% NOTE: This predicate works correctly under the assumption that the 
%%% current Node only includes implications, abducible atoms, and no
%%% defined atoms.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

getAbducibles( [], []).
getAbducibles( [provable(_)|Tail], List) :-
	!, getAbducibles( Tail, List).
getAbducibles( [_ implies _|Tail], List) :-
	!, getAbducibles( Tail, List).	
getAbducibles( [_ / _|Tail], List) :-
	!, getAbducibles( Tail, List).	
getAbducibles( [_ = _|Tail], List) :-
	!, getAbducibles( Tail, List).
getAbducibles( [Head|Tail], List) :-
	is_constraint( Head), 
	!, 
	getAbducibles( Tail, List).	
getAbducibles( [Head|Tail], [Head|NList]) :-
	!, getAbducibles( Tail, NList).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% getLiterals( +Node, -Literals).
%%% Literals is a list of all the literals from a node Node.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

getLiterals( [], []).
getLiterals( [provable(A)|Tail], List) :-
	!, 
	getLiterals( Tail, List1),
	append( A, List1, List).
getLiterals( [A implies B|Tail], List) :-
	!, 
	getLiterals( A, List1),
	getLiterals( B, List2),
	getLiterals( Tail, List3),
	append( List1, List2, List4),
	append( List4, List3, List).	
getLiterals( [_ / _|Tail], List) :-
	!, getLiterals( Tail, List).
getLiterals( [_ = _|Tail], List) :-
	!, getLiterals( Tail, List).
getLiterals( [false|Tail], List) :-
	!, getLiterals( Tail, List).
getLiterals( [*|Tail], List) :-
	!, getLiterals( Tail, List).
getLiterals( [Head|Tail], List) :-
	is_constraint( Head), 
	!, 
	getLiterals( Tail, List).	
getLiterals( [Head|Tail], [Head|NList]) :-
	!, getLiterals( Tail, NList).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% removeDefs( +Literals, +Defs, -List).
%%% List is a list all the elements from Literals list except those 
%%% that also appear in Defs.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

removeDefs( [], _, []).
removeDefs( [A|T1], Defs, T2) :-
	get_def( A, Defs, _),
	removeDefs( T1, Defs, T2).
removeDefs( [A|T1], Defs, [A|T2]) :-
	\+ get_def( A, Defs, _),
	removeDefs( T1, Defs, T2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% random_variables( +List, -NList).
%%% NList is a copy of List with all the elements of List copied.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

random_variables( [], []).
random_variables( [A|T1], [B|T2]) :-
	copy_term( A:_, B:_),
	random_variables( T1, T2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% getFalseAbducibles( +Abducibles, +AllAbducibles, -FalseAbducibles).
%%% FalseAbducibles is a list of all the abducibles that must be 
%%% marked as false.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

getFalseAbducibles( [], [], []).
getFalseAbducibles( [], List, List).
getFalseAbducibles( [A|T1], T2, List) :-
	getFalseAbducibles2( A, T2, T22),
	getFalseAbducibles( T1, T22, List).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% getFalseAbducibles( +Abducible, +Abducibles, -FalseAbducibles).
%%% FalseAbducibles is the same as Abducibles list without any element
%%% of the form Abducible.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

getFalseAbducibles2( _, [], []).
getFalseAbducibles2( A, [A|T1], T3) :-
	getFalseAbducibles2( A, T1, T3).
getFalseAbducibles2( A, [B|T1], [B|T3]) :-
	A \= B,
	getFalseAbducibles2( A, T1, T3).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% getFalseAbducibles( +List, -FalseList).
%%% FalseList is the same as List with every variable H in List
%%% substituted with ``H iff [[false]]''.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

createFalseList( [], []).
createFalseList( [H|T1], [H iff [[false]]|T2]) :-
	createFalseList( T1, T2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% getDefsList( +Defs, -List).
%%% Defs has elements of the form _ iff L and List is all the L's 
%%% put together into a list.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

getDefsList( [], []).
getDefsList( [_ iff L|Tail], T2) :-
	listOfListToList( L, L1),
	getDefsList( Tail, T1),
	append( L1, T1, T2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% listOfListToList( +ListOfLists, -List).
%%% List is the all the lists in the ListOfLists appended together.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

listOfListToList( [], []).
listOfListToList( [[]], []).
listOfListToList( [H1|T1], T2) :-
	listOfListToList( T1, T22),
	append( H1, T22, T2).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
